/* Version 1.0 (Beta Test)
 * Copyright (c) Rui Zhang, 2007
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The code will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - All bug fixes will be returned to rui@csse.unimelb.edu.au for
 *    inclusion in future releases
 */

/*************************************************************************
*
* File : btree.c
*
* Modules:
* Routines to manipulate a B+ tree
*   int BT_Open()
*   int BT_Close()
*   int BT_Insert()
*   int BT_Search()
*   int BT_Search_v2()
*   int BT_SearchRange()
*
* Version 1
* Coded by: Dai Haoyu		Modified by:
* Date: 12/9/97			Date modified: 
*************************************************************************/
#include "btree.h"
#include <float.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

PTR curleaf, tail;

/* function prototype */
/* internal */
int BT_Open(BTREE *, FILE *, PTR, PTR);
int BT_Close(BTREE *);
int BT_Insert(BTREE *, const long, const KEYTYPE, RECORD_TYPE *);
int BT_Search(BTREE *, PTR, RECORD_TYPE *, const KEYTYPE);
PTR BT_Locate(BTREE *, PTR, const KEYTYPE);
/*int BT_Search_v2(BTREE *, PTR, const KEYTYPE);*/
int BT_SearchRange(BTREE *, PTR, const KEYTYPE, const KEYTYPE, range *);
void refine(BTREE *);
void compact(BTREE *btree);
int BuildINodes(PTR ptr, BTREE *btree, KEYTYPE *keymin, KEYTYPE *keymax);
IMINODE* LoadINodes(PTR ptr, BTREE *btree);
PTR IMBT_Locate(IMINODE *, KEYTYPE);
IMINODE* IMBT_LocateBaseINode(IMINODE *piminode, KEYTYPE key);

/* external */
extern int BT_NewINode(INODE **);
extern int BT_NewLNode(LNODE **);
extern PTR BT_SelectChild(INODE *, KEYTYPE);
extern int BT_InsertNewINode(INODE *, PTR, KEYTYPE, PTR);
extern PTR BT_IOWriteNewINode(BTREE *, INODE *);
extern int BT_IOWriteINode(BTREE *, PTR, INODE *);
extern int BT_IOReadINode(BTREE *, PTR, INODE *);
extern int BT_InsertR(BTREE *, const long, const KEYTYPE, RECORD_TYPE *, INODE *, 
                      INODE **, KEYTYPE *);
extern void JoinList(OFFSETLIST *, OFFSETLIST *, OFFSETLIST **);
extern int BT_SearchLeaf(BTREE *, PTR, RECORD_TYPE *, const KEYTYPE,OFFSETLIST **);
/*extern int BT_SearchLeaf_v2(BTREE *, PTR, const KEYTYPE, OFFSETLIST **);*/
extern int BT_SearchLeafRange(BTREE *, PTR, const KEYTYPE, const KEYTYPE, range *,
                              OFFSETLIST **);
extern int BT_InsertSrhResult(BTREE *, OFFSETLIST *);

extern int BT_VerifyINode(INODE *);
extern void BT_PrintINode(INODE *);
extern void DeleteList(OFFSETLIST **);
extern int IOClose(FILE *);

extern int BT_IOReadLNode(BTREE *, PTR, LNODE *);
extern int BT_IOWriteLNode(BTREE *, PTR, LNODE *);

/***************************************************************************

* BT_Open():
*   Routine to open a BTree
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TURE if opened successfully else FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   The BTREE struct is properly initialsed
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_Open(BTREE *btree,  /* the BTREE struct to be initialised */
            FILE *fp,      /* the file containing the B+ tree  */
            PTR  root,    /* the offset of the root node in the file */
            PTR  lhead    /* the offset of the leaf head node in the file */
) {
  int result = TRUE;

  btree->BTreeFile = fp;
  btree->Root = root;
  btree->Leaf_Head = lhead;
/**/
  btree->top=-1;

  return result;
}

/***************************************************************************

* functionName:
*   ShortDescription
*
* IMPORTS:
*   None.
*
* RETURNS:
*   None.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_Close(BTREE *btree  /* the BTREE struct of which the B+ tree
                            * is to be closed */
) {
  int result = FALSE;

  if (IOClose(btree->BTreeFile)){
    result = TRUE;
  }

  return result;
}

/***************************************************************************

* BT_Insert():
*   Routine to insert a new record to the B+ tree
*
* IMPORTS:
*   BT_NewINODE()
*   BT_IOWriteNewINode()
*   BT_IOWriteINode()
*   BT_IOReadINode()
*   BT_InsertR()
*   BT_INsertNewINode()
*
* RETURNS:
*   TRUE if insertion is successful else FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   The B+ tree index file will be updated accordingly
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_Insert(BTREE *btree,      /* the btree to insert the record */
              const long recid,  /* the offset of record in the data file */
              //float key,
              const KEYTYPE key,  /* the key value of the reocord */
              RECORD_TYPE *rd
) { 
  int result = FALSE;
  INODE *root,
        *newNode,
        *newRoot;
  PTR offset;
  KEYTYPE middleKey;  /*the key value to be moved up when splitting occurs*/

  if (!BT_NewINode(&root)) {
    printf("BT_Insert : Can't allocate memory for root node \n");
    return result;
  }

  if (btree->Root <0){  /* index is empty */
    root->Level = 0;
    root->NumKey = 0;
    btree->Root = BT_IOWriteNewINode(btree, root);
    printf("Root of b+ tree is created\n");
  }

  if (!BT_IOReadINode(btree, btree->Root, root)) {
    printf("BT_Insert : Can't read from index file\n");
    return result;
  }

  if (BT_InsertR(btree, recid, key, rd, root, &newNode, &middleKey)) {
  /* no split occurs */
    if (!BT_IOWriteINode(btree, btree->Root, root)) {
      printf("BT_Insert : Can't write to index file\n");
      return result;
    } else {
     free(root);
    }
  } else {  /* spliting occurs */
    printf("Splitting at root level\n");  

    if (!BT_VerifyINode(root)) {
      BT_PrintINode(root);
    }
    if (!BT_VerifyINode(newNode)) {
      BT_PrintINode(newNode);
    }
    BT_IOWriteINode(btree, btree->Root, root);
    BT_NewINode(&newRoot);
    newRoot->Level = root->Level + 1;
    newRoot->NumKey = 0;
    offset = BT_IOWriteNewINode(btree, newNode);
/**/
    printf("************ Note: Offset of new root :  %d",btree->Root);

    BT_InsertNewINode(newRoot, btree->Root, middleKey, offset);
    btree->Root = BT_IOWriteNewINode(btree, newRoot);
    printf("offset of new root%d",btree->Root);

    free(root);
    free(newRoot);
    free(newNode);
  }

  result = TRUE;
  return result;
}



PTR BT_Locate( BTREE *btree,
               PTR currNodePtr,   /* the offset of the current node */          
               const KEYTYPE key  /* the key to be searched */    
) {

  int i;
  INODE *currNode;
  PTR child, lchild;
  
  if (currNodePtr < 0L) {
    return -1;
  }
  
  BT_NewINode(&currNode);

  if (BT_IOReadINode(btree, currNodePtr, currNode)){
     if (currNode->Level > 0) {  /* not a leaf level INODE */
        child = BT_SelectChild(currNode, key);
        lchild = BT_Locate(btree, child, key);
     } else {  /* a leaf level INODE , the next level is LNODE */
        lchild = BT_SelectChild(currNode, key);      
     }
  }
  free(currNode);
  return (lchild);  

}


/***************************************************************************
* BT_Search():
*   Routine to search a record in the index
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if searching is successful else FALSE.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   The set of record ids of found record will be passed as an offsetlist
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_Search(BTREE *btree,     /* the B+ tree to searched from */
              PTR currNodePtr,   /* the offset of the current node */
              RECORD_TYPE *rd,
              const KEYTYPE key  /* the key to be searched */
) {
  int result = TRUE;
  int i;
  INODE *currNode;
  PTR child;
  OFFSETLIST *subList = NULL;  

  if (currNodePtr < 0L) {
    return result;
  }
  
  BT_NewINode(&currNode);

  if (BT_IOReadINode(btree, currNodePtr, currNode)){
    if (currNode->Level > 0) {  /* not a leaf level INODE */

/* alternative 2 : to invoke BT_Search only for the left-most qualifing
 *                 child and perform sequential scan at LNODE level */
      child = BT_SelectChild(currNode, key);
      BT_Search(btree, child, rd, key);

    } else {  /* a leaf level INODE , the next level is LNODE */
/* alternative 2 */
      child = BT_SelectChild(currNode, key);
      BT_SearchLeaf(btree, child, rd, key,&subList);
      BT_InsertSrhResult(btree,subList);
      DeleteList(&subList);
/* alternative 2 */
    }
  }

  free(currNode);
  return (result);
}



/***************************************************************************

* BT_SearchRange():
*   Routine to search a range in the index
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if searching is successful else FALSE.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   The set of record ids of found record will be passed as an offsetlist
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_SearchRange(BTREE *btree,     /* the B+ tree to searched from */
              PTR currNodePtr,   /* the offset of the current node */
              const KEYTYPE key1,  /* lower bound of the range */
              const KEYTYPE key2,   /* upper bound of the range */
              range *rg
) {
  int result = TRUE;
  int i;
  INODE *currNode;
  OFFSETLIST *subList = NULL;
  PTR child;

  if (currNodePtr < 0L) {
    return result;
  }
  
  BT_NewINode(&currNode);

  if (BT_IOReadINode(btree, currNodePtr, currNode)){
    if (currNode->Level > 0) {  /* not a leaf level INODE */      

/* alternative 2 : to invoke BT_Search only for the left-most qualifing
 *                 child and perform sequential scan at LNODE level */
      child = BT_SelectChild(currNode, key1);
      BT_SearchRange(btree, child, key1, key2, rg);

    } else {  /* a leaf level INODE , the next level is LNODE */
/* alternative 2 */
      subList = NULL;
      child = BT_SelectChild(currNode, key1);
      BT_SearchLeafRange(btree, child, key1, key2, rg, &subList);
      BT_InsertSrhResult(btree, subList);
      DeleteList(&subList);
/* alternative 2 */
    }
  }

  free(currNode);
  return (result);
}





/********************************************************************/
/*only for KNN to add previus pointer between leafnodes.*/


void refine(BTREE *btree)   
{

   PTR pointer, prev; 
   LNODE *lNode=NULL;
   int Continue=1;
   
   BT_NewLNode(&lNode);        
   pointer=btree->Leaf_Head;
   
   prev=-1; 
   while (Continue) {       
      BT_IOReadLNode(btree, pointer, lNode);          
      lNode->Prev=prev;
      BT_IOWriteLNode(btree, pointer, lNode);
      prev=pointer;
      pointer=lNode->Next;
      if (pointer<0) {
         Continue=0;
      } else { 
      }
   }
   free(lNode);
}

/********************************************************************/

void compact(BTREE *btree)
{
	PTR addr1, addr2, oriaddr1, fromover, toover, tonext, leafhead, treeroot; 
	LNODE *lNode1, *lNode2;
	ONODE *oNode1, *oNode2;
	int to, from, i, j, flag, tonodetype, fromnodetype;
	int clfanout;
	KEYTYPE x,y;
	
	if(compactrate<0.75)
	{
		printf("\nError!: Compacting b-tree failes. Please specify a larger utilization rate.\n\n");
		return;
	}
	
	clfanout=(int)(compactrate * L_FAN_OUT);
	
	BT_NewLNode(&lNode1);
	BT_NewLNode(&lNode2);
	BT_NewONode(&oNode1);
	BT_NewONode(&oNode2);
	addr1=leafhead=btree->Leaf_Head;
	treeroot=btree->Root;
	
	//find the first node with less than clfanout entries
	do
	{
		BT_IOReadLNode(btree, addr1, lNode1);
		oriaddr1=addr1;
		addr1=lNode1->Next;
		to=lNode1->NumEntry;
	}while(to==clfanout-1 && addr1>=0);
	
	toover=lNode1->Overflow;
	tonext=addr2=addr1; //addr2 points to the next node of lNode1;
	from=0;
	
	flag=1;
	tonodetype=fromnodetype=1; //1 for leaf; 0 for overflow node.
	
	if(addr2>=0) BT_IOReadLNode(btree, addr2, lNode2);
	else flag=0;
	
	while (flag>0)
	{
		if(fromnodetype)
		{
			for( ; to < clfanout && from < lNode2->NumEntry ; to++, from++)
				lNode1->Entries[to]=lNode2->Entries[from];
		}
		else
		{
			for( ; to < clfanout && from < oNode2->NumEntry ; to++, from++)
				lNode1->Entries[to]=oNode2->Entries[from];
		}
		
		if(fromnodetype && from==lNode2->NumEntry)
		{
			from=0;
			fromover=lNode2->Overflow;
			addr2=lNode2->Next;
			
			if(fromover>=0)
			{
				BT_IOReadONode(btree,fromover,oNode2);
				fromnodetype=0;
			}
			else if(addr2>=0)
			{
				BT_IOReadLNode(btree, addr2, lNode2);
			}
			else flag=0;
		}
		else if(!fromnodetype && from==oNode2->NumEntry)
		{
			from=0;
			fromover=oNode2->Next;
			if(fromover>=0)
			{
				BT_IOReadONode(btree,fromover,oNode2);
			}
			else if(addr2>=0)
			{
				BT_IOReadLNode(btree, addr2, lNode2);
				fromnodetype=1;
			}
			else flag=0;
		}
		
		if(to==clfanout)
		{
			if(!flag) 
			{
				lNode1->NumEntry=to;
				lNode1->Next=-1;
				lNode1->Overflow=-1;
				lNode1->overflowNum=0;
				BT_IOWriteLNode(btree, oriaddr1, lNode1);
				flag=-1;
			}
			
			else
			{
				if(toover>=0)
				{
					lNode1->NumEntry=to;
					lNode1->Next=toover;
					lNode1->Overflow=-1;
					lNode1->overflowNum=0;					
					BT_IOWriteLNode(btree, oriaddr1, lNode1);

					BT_IOReadONode(btree,toover,oNode1);
					oriaddr1=toover;
					toover=oNode1->Next;
					tonodetype=0;
				}
				else if(tonext>=0)
				{
					lNode1->NumEntry=to;
					lNode1->Next=tonext;
					lNode1->Overflow=-1;
					lNode1->overflowNum=0;					
					BT_IOWriteLNode(btree, oriaddr1, lNode1);

					BT_IOReadLNode(btree, tonext, lNode1);
					oriaddr1=tonext;
					toover=lNode1->Overflow;
					tonext=lNode1->Next;
					tonodetype=1;
				}
				else flag=0;
				
				to=0;
			}
		}
	}
	
	if(flag != -1)
	{
		lNode1->NumEntry=to;
		lNode1->Next=-1;
		lNode1->Overflow=-1;
		lNode1->overflowNum=0;
		BT_IOWriteLNode(btree, oriaddr1, lNode1);
	}
	
	curleaf=btree->Leaf_Head;
	BuildINodes(treeroot, btree, &x, &y);

	free(lNode1);
	free(lNode2);
	free(oNode1);
	free(oNode2);
}


/********************************************************************/
int BuildINodes(PTR ptr, BTREE *btree, KEYTYPE *keymin, KEYTYPE *keymax)
{
	int i, j, flag;
	PTR p;
	int cifanout;
	INODE *pinode, *pchild;
	LNODE *lNode;
	KEYTYPE curkeymin, curkeymax, childkeymin, childkeymax;
	
	cifanout=(int)(compactrate * I_FAN_OUT);
	
	BT_NewINode(&pinode);
	BT_IOReadINode(btree, ptr, pinode);
	
	BT_NewINode(&pchild);
	BT_NewLNode(&lNode);

	flag=1;

	curkeymin=FLT_MAX;
	curkeymax=FLT_MIN;

	if(pinode->Level>0)
	{
		for(i=0;i<pinode->NumKey+1 && flag>0;i++)
		{
			flag=BuildINodes(pinode->Children[i], btree, &childkeymin, &childkeymax);
			if(flag>=0)
			{
				if(i==0) curkeymin=childkeymin;
				else pinode->Keys[i-1]=childkeymin;		
			}
		}
		
		*keymin=curkeymin;
		pinode->NumKey=i-1;
		
		BT_IOWriteINode(btree, ptr, pinode);
		
		free(pinode);
		free(pchild);
		free(lNode);

		if(i==1 && flag<0) return -1;//this inode has no child inode
		else if(flag==0) return 0;
		else return 1;					
	}
	
	if(pinode->Level==0)
	{
		if(curleaf>=0)
		{
			BT_IOReadLNode(btree, curleaf, lNode);
			pinode->Children[0]=curleaf;
			pinode->NumKey=0;
			curleaf=lNode->Next;
			*keymin=lNode->Entries[0].Key;
		}
		else 
		{
			free(pinode);
			free(pchild);
			free(lNode);
			return -1;
		}
			
		for( i=0; curleaf>=0 && i<cifanout-1; i++ )
		{
			BT_IOReadLNode(btree, curleaf, lNode);
			pinode->Children[pinode->NumKey+1]=curleaf;
			pinode->Keys[pinode->NumKey]=lNode->Entries[0].Key;
			pinode->NumKey++;
			curleaf=lNode->Next;
		}
		
		BT_IOWriteINode(btree, ptr, pinode);

		free(pinode);
		free(pchild);
		free(lNode);		
		
		if(curleaf<0) return 0;
		else return 1;
	}
}

/********************************************************************/
IMINODE* LoadINodes(PTR ptr, BTREE *btree)
{
	int i;
	INODE *pinode, *pbaseiminode;
	IMINODE *piminode;

	
	BT_NewINode(&pinode);
	BT_IOReadINode(btree, ptr, pinode);
	
	if(pinode->Level>0)
	{
		piminode=(IMINODE *)malloc(sizeof(IMINODE));
		piminode->Level=pinode->Level;
		piminode->NumKey=pinode->NumKey;
		for(i=0;i<pinode->NumKey;i++) piminode->Keys[i]=pinode->Keys[i];
		for(i=0;i<=pinode->NumKey;i++) 
		{piminode->Children[i]=LoadINodes(pinode->Children[i], btree);}
		
		free(pinode);

		return piminode;
	}
	
	else if(pinode->Level==0)
	{
		pbaseiminode=(INODE *)malloc(sizeof(INODE));
		pbaseiminode->Level=pinode->Level;
		pbaseiminode->NumKey=pinode->NumKey;
		for(i=0;i<pinode->NumKey;i++) pbaseiminode->Keys[i]=pinode->Keys[i];
		for(i=0;i<=pinode->NumKey;i++)
		{ pbaseiminode->Children[i]=pinode->Children[i];}
		
		free(pinode);
		
		return pbaseiminode;
	}
	
	else
	{
		free(pinode);
		printf("Load INodes Error\n");
		exit(0);
	}
}


/********************************************************************/

PTR IMBT_Locate(IMINODE *piminode, KEYTYPE key)
{
	int i;
	INODE *pinode;
	pinode=(INODE *)IMBT_LocateBaseINode(piminode,key);
	
	//scan the base inode to locate the page to be read
	if(pinode->NumKey==0) return -1;
	else
	{
		if(key<pinode->Keys[0])	return pinode->Children[0];
		
		for(i=0;i<pinode->NumKey-1;i++)
		{
			if(pinode->Keys[i]<=key && key<pinode->Keys[i+1]) return pinode->Children[i+1];
		}
		
		return pinode->Children[pinode->NumKey];
	}
	//End scan the base inode to locate the page to be read
}

/********************************************************************/

IMINODE* IMBT_LocateBaseINode(IMINODE *piminode, KEYTYPE key)
{
	int i;
	
	if(piminode->Level==0) return piminode;
	
	if(piminode->Level==1)
	{
		if(key<piminode->Keys[0])	return piminode->Children[0];
		
		for(i=0;i<piminode->NumKey-1;i++)
		{
			if(piminode->Keys[i]<=key && key<piminode->Keys[i+1]) return piminode->Children[i+1];
		}
		
		return piminode->Children[piminode->NumKey];
	}
	
	if(piminode->Level>1)
	{
		if(key<piminode->Keys[0])	return IMBT_LocateBaseINode(piminode->Children[0],key);
		
		for(i=0;i<piminode->NumKey-1;i++)
		{
			if(piminode->Keys[i]<=key && key<piminode->Keys[i+1]) return IMBT_LocateBaseINode(piminode->Children[i+1],key);
		}
		
		return IMBT_LocateBaseINode(piminode->Children[piminode->NumKey],key);
	}
}

DATA_TYPE Dis_Value(RECORD_TYPE record) {

   int i, j, k;
   float p[D];
   DATA_TYPE value, distance[2*D], dd;
   
   for (i=0; i<D; i++) {
      for (j=0; j<D; j++) {
         if (j==i) {
           p[j]=0.0;  
         } else {
           p[j]=0.5;
         }
      }
      dd=0.0;
      for (k=0;k<D;k++) {
           /*printf("**** %f   %f    %f     %f\n",record.fields[k],p[k],record.fields[k]-p[k],(record.fields[k]-p[k])*(record.fields[k]-p[k]));
*/
           dd+=(record.fields[k]-p[k])*(record.fields[k]-p[k]);     
      }          
      distance[i] = pow(dd,0.5);   
     /* printf("%d   %lf\n",i,distance[i]); */
   }   
   
   for (i=D; i<2*D; i++) {
      for (j=0; j<D; j++) {
         if (j==i-D) {
            p[j]=1.0;
         } else {
            p[j]=0.5;
         }         
      }
      dd=0.0;
      for (k=0;k<D;k++) {
           /*printf("**** %f   %f    %f     %f\n",record.fields[k],p[k],record.fields[k]-p[k],(record.fields[k]-p[k])*(record.fields[k]-p[k]));
*/         
          dd+=(record.fields[k]-p[k])*(record.fields[k]-p[k]);     
      }          
      distance[i]=pow(dd,0.5);    
    /*  printf("%d   %lf\n",i,distance[i]); */
   }
   k=0;
   value=distance[0];
   for (i=1; i<2*D; i++) {
      if (value > distance[i]) {
          value = distance[i];
          k=i;
      }
   } 
   
   /* printf("key value= %d*10+%lf = ",k,value); */  
   value = k*10.0 + value;
  /* printf("%lf \n \n", value);*/
   
   return value;
}
